"use strict";
/*************************************************************************
 *
 * Troven CONFIDENTIAL
 * __________________
 *
 *  (c) 2017-2020 Troven Pty Ltd
 *  All Rights Reserved.
 *
 * NOTICE:  All information contained herein is, and remains
 * the property of Troven Pty Ltd and its licensors,
 * if any.  The intellectual and technical concepts contained
 * herein are proprietary to Troven Pty Ltd
 * and its suppliers and may be covered by International and Regional Patents,
 * patents in process, and are protected by trade secret or copyright law.
 * Dissemination of this information or reproduction of this material
 * is strictly forbidden unless prior written permission is obtained
 * from Troven Pty Ltd.
 */
Object.defineProperty(exports, "__esModule", { value: true });
const _ = require("lodash");
const with_db_1 = require("mime-type/with-db");
const fs = require("fs");
const api_service_core_1 = require("api-service-core");
// const request = require("request");
const Handlebars = require("handlebars");
/**
 * Textual Interpolation
 * -------------
 * Render a Handlebars template based on static meta-data and optional dynamic API call.
 *
 * The same JWT used for this UI layer will be forwarded to the API endpoint (unless reuseJWT = true).
 *
 * Bindings can be injected from config or options -
 *      static:
 *      brand:
 *          title: Site / App name
 *          description: Generic site description
 *      apis:
 *          my_app:
 *              url: request-url
 *              headers: request-headers
 *              etc
 *      feature:
 *          a: feature toggle A
 *          b: toggle B
 *
 * @type {{name: string, title: string, description: string, _render: module.exports._render, fn: module.exports.fn}}
 */
class textual {
    constructor() {
        this.name = "textual";
        this.title = "Server-side text-based transformation";
    }
    install(context, _options) {
        let options = _.extend({
            mimetypes: {
                "text/n3": {
                    extensions: "n3",
                },
            },
        }, _options);
        _.each(options.mimetypes, (opt, type) => {
            with_db_1.default.define(type, opt);
        });
        context.middleware.set(new TextualFeature(context, options));
    }
}
exports.default = textual;
class TextualFeature {
    constructor(context, options) {
        this.name = "textual";
        this.context = context;
        this.options = options;
    }
    /**
     * Collect a set of API responses as promise
     *
     * @param apis
     * @param options
     * @param token
     */
    fn(op, _options) {
        let options = _.extend({ vars: {} }, this.options, _options);
        let context = op.context;
        let apis = _.extend({}, this.options.apis, options.apis);
        if (options.api)
            apis.data = options.api;
        let global_vars = _.extend({}, this.options.vars, _options.vars);
        context.log({
            code: "plugin:textual",
            resource: op.resource,
            reuseJWT: options.reuseJWT,
            apis: _.keys(apis),
        });
        // renderer middleware
        return (req, res, next) => {
            let jwt_token = options.reuseJWT ? req.header("authorization") : "";
            let op_ctx = api_service_core_1.APIs.getOperationContext(op, req, global_vars);
            // load and interpolate a template file
            let filename = api_service_core_1.Vars.$(options.filename, op_ctx);
            let raw_template = null;
            try {
                raw_template = fs.readFileSync(filename, "utf8");
            }
            catch (err) {
                context.warn({
                    code: "plugin:textual:file:missing",
                    filename: filename,
                    error: err.message,
                });
                return next();
            }
            let template = Handlebars.compile(raw_template);
            let file_mime = with_db_1.default.lookup(filename);
            context.log({
                code: "plugin:textual:apis",
                apis: _.keys(apis),
                filename: filename,
                mimetype: file_mime,
            });
            api_service_core_1.APIs.requests(this.context, apis, op_ctx, jwt_token)
                .then((results) => {
                let render_context = _.extend({}, op_ctx, results);
                let content = template(render_context);
                res.status(200);
                res.header("content-type", file_mime);
                res.send(content);
            })
                .catch((err) => {
                console.error("plugin:textual:apis:failed: %j", err.message);
                // render error response
                let _error = {
                    code: "plugin:textual:apis:failed",
                    filename: filename,
                    message: err.message,
                    error: err.message,
                };
                context.error(_error);
                res.status(500);
                res.send({
                    code: "plugin:textual:apis:failed",
                    request: req.a6s_uuid,
                });
            });
        };
    }
}
//# sourceMappingURL=textual.js.map