import * as _ from "lodash";
import * as fileupload from "express-fileupload";

import {
    IChassisContext,
    IChassisPlugin,
    IChassisMiddleware,
    IOperation,
    Vars,
    APIs,
} from "api-service-core";
import { Handler } from "express-serve-static-core";
// import { assert } from "console";
import { writeFileSync, readdirSync, readFileSync, unlinkSync } from "fs";
import * as mkdirp from "mkdirp";
import assert = require("assert");

/**
 * plugin to handle JSON body
 *
 * @param context
 * @param options
 * @returns {exports}
 */
export default class BlobsPlugin implements IChassisPlugin {
    name = "blobs";
    title = "support for folders of binary objects";
    context: IChassisContext;
    options: any;
    mimetypes: [string];

    install(context: IChassisContext, _options: any) {
        this.context = context;
        this.mimetypes = ["*"];
        this.options = _.extend(
            {
                folder: "",
                tempFileDir: "tmp",
                useTempFiles: true,
                createParentPath: true,
                safeFileNames: true,
                preserveExtension: 6,
                abortOnLimit: true,
                parseNested: true,
                limits: { fileSize: 1024 * 1024 * 1024 }, // 1gb
            },
            _options
        );

        assert(this.options.folder, "missing {{blobs.folder}}");

        context.api.use(fileupload(this.options));

        context.middleware.set(new BlobsMiddleware(this));
    }
}

class BlobsMiddleware implements IChassisMiddleware {
    name: string = "blobs";
    title?: string = "support for file-based binary objects";
    defaults?: any = {};

    constructor(protected plugin: BlobsPlugin) {}

    fn(oper: IOperation, _options: any): Function | Handler {
        assert(_options.path, "missing {{blobs.paths}}");

        return (req: any, res: any, _next: Function) => {
            let ctx = APIs.getOperationContext(oper, req);
            let file_path = Vars.$(
                this.plugin.options.folder + "/" + _options.path,
                ctx
            );
            let options = _.extend({}, _options, { path: file_path });
            mkdirp.sync(file_path + "/meta");
            mkdirp.sync(file_path + "/blob");

            switch (oper.actionId) {
                case "post":
                    this.uploads(req, res, options);
                    break;
                case "get":
                    if (oper.resource.endsWith("/"))
                        this.list(req, res, options);
                    else this.download(req, res, options);
                    break;
                case "put":
                    this.update(req, res, options);
                    break;
                case "delete":
                    this.delete(req, res, options);
                    break;
            }
            this.plugin.context.log({
                code: "api:blob:action:" + oper.actionId,
                path: options.path,
            });
        };
    }

    list(_req: any, res: any, options: any) {
        let meta_folder = options.path + "/meta/";
        let meta_files = readdirSync(meta_folder);
        let files = [];
        _.each(meta_files, (file) => {
            let json = JSON.parse(readFileSync(meta_folder + file, "UTF-8"));
            files.push(json);
        });
        this.plugin.context.log({
            code: "api:blob:list",
            path: meta_folder,
            files: files.length,
        });

        res.status(200).send(files);
    }

    meta(_req: any, options: any) {
        let id = this.get_filename(_req.params._id);
        if (!id) return null;
        let meta_file = options.path + "/meta/" + id + ".json";
        try {
            let meta = JSON.parse(readFileSync(meta_file, "UTF-8"));
            let blob_file = options.path + "/blob/" + id + "." + meta.type;
            return {
                _id: id,
                meta: meta,
                blob_file: blob_file,
                meta_file: meta_file,
            };
        } catch (e) {
            this.plugin.context.log({
                code: "api:blob:meta:error",
                path: meta_file,
                cwd: process.cwd(),
                error: e.message,
            });
            return null;
        }
    }

    sendError(status: number, _res) {
        _res.status(status).end();
    }

    download(_req: any, _res: any, options: any) {
        let meta = this.meta(_req, options);
        if (!meta) return this.sendError(404, _res);
        this.plugin.context.log({
            code: "api:blob:download",
            path: meta.blob_file,
            cwd: process.cwd(),
        });
        _res.status(200).sendFile(meta.blob_file, { root: process.cwd() });
    }

    update(_req: any, _res: any, options: any) {
        this.uploads(_req, _res, options);
    }

    delete(_req: any, _res: any, options: any) {
        let meta = this.meta(_req, options);
        if (!meta) return this.sendError(404, _res);
        try {
            unlinkSync(meta.meta_file);
            unlinkSync(meta.blob_file);
            this.plugin.context.log({
                code: "api:blob:deleted",
                path: meta.blob_file,
                cwd: process.cwd(),
            });
            _res.status(200);
            _res.send();
        } catch (e) {
            this.plugin.context.log({
                code: "api:blob:delete:failed",
                path: meta.blob_file,
                error: e.message,
            });
            _res.status(400);
            _res.send();
        }
    }

    asserFiles(req: any, res: any) {
        if (!req.files || Object.keys(req.files).length === 0) {
            let msg = { code: "api:blob:action:post:missing-files" };
            this.plugin.context.error(msg);
            res.status(400).send(msg);
            return false;
        }
        return true;
    }

    isAllowed(mimetype: string) {
        if (!this.plugin.mimetypes) return true;
        if (this.plugin.mimetypes.indexOf("*") >= 0) return true;
        console.log("is-allowed: %j -> %j", mimetype, this.plugin.mimetypes);
        return this.plugin.mimetypes.indexOf(mimetype) >= 0;
    }
    uploads(req: any, res: any, options: any) {
        if (!this.asserFiles(req, res)) return;

        let uploaded = [];
        let fails = 0;
        _.each(req.files, (file) => {
            let meta = {
                _id: file.md5,
                filename: file.name,
                mimetype: file.mimetype,
                size: file.size,
                type: this.get_extension(file.name),
                error: null,
            };
            if (this.upload(file, meta, options)) {
                uploaded.push(meta);
            } else {
                fails++;
                meta.error = {
                    code: "not-allowed",
                    filename: file.name,
                    mimetype: file.mimetype,
                    allowed: this.plugin.mimetypes,
                };
                uploaded.push(meta);
                this.plugin.context.warn({
                    code: "api:blob:uploaded:not-allowed",
                    path: options.path,
                    filename: file.name,
                    mimetype: file.mimetype,
                    allowed: this.plugin.mimetypes,
                });
            }
        });
        res.status(fails ? 415 : 201);
        res.send(uploaded);
    }

    upload(file: any, meta: any, options: any) {
        if (!this.isAllowed(meta.mimetype)) return false;

        let meta_filename = options.path + "/meta/" + meta._id;
        writeFileSync(meta_filename + ".json", JSON.stringify(meta));

        let blob_filename =
            options.path + "/blob/" + meta._id + "." + meta.type;
        file.mv(blob_filename).then( (saved) => {
          this.plugin.context.log({
            code: "api:blob:uploaded",
            path: options.path,
            uuid: meta._id,
            saved: saved || false,
            file: file || false,
            filename: blob_filename,
          });

          this.plugin.context.bus.emit("api:blob:uploaded", {
            meta: meta,
            options: options,
            from_filename: file.tempFilePath,
            to_filename: blob_filename,
          });

        }).catch(err => {
          this.plugin.context.log({
            code: "api:blob:upload:failed",
            path: options.path,
            uuid: meta._id,
            filename: blob_filename,
            error: err.message
          });

        })
        return true; // "means 'is allowed' .. not 'upload successful'"
    }

    get_extension(name) {
        let ix = name.lastIndexOf(".");
        if (ix < 0) return null;
        return name.substring(ix + 1);
    }

    get_filename(name) {
        let ix = name.lastIndexOf(".");
        if (ix < 0) return name;
        return name.substring(0, ix);
    }
}
