/*************************************************************************
 *
 * Troven CONFIDENTIAL
 * __________________
 *
 *  (c) 2017-2020 Troven Pty Ltd
 *  All Rights Reserved.
 *
 * NOTICE:  All information contained herein is, and remains
 * the property of Troven Pty Ltd and its licensors,
 * if any.  The intellectual and technical concepts contained
 * herein are proprietary to Troven Pty Ltd
 * and its suppliers and may be covered by International and Regional Patents,
 * patents in process, and are protected by trade secret or copyright law.
 * Dissemination of this information or reproduction of this material
 * is strictly forbidden unless prior written permission is obtained
 * from Troven Pty Ltd.
 */

import * as _ from "lodash";
import mimetypes from "mime-type/with-db";
import * as fs from "fs";
import {
    IChassisContext,
    IChassisPlugin,
    IChassisMiddleware,
    IOperation,
    APIs,
    APIRequests,
    Vars,
} from "api-service-core";

// const request = require("request");
const Handlebars = require("handlebars");

/**
 * Textual Interpolation
 * -------------
 * Render a Handlebars template based on static meta-data and optional dynamic API call.
 *
 * The same JWT used for this UI layer will be forwarded to the API endpoint (unless reuseJWT = true).
 *
 * Bindings can be injected from config or options -
 *      static:
 *      brand:
 *          title: Site / App name
 *          description: Generic site description
 *      apis:
 *          my_app:
 *              url: request-url
 *              headers: request-headers
 *              etc
 *      feature:
 *          a: feature toggle A
 *          b: toggle B
 *
 * @type {{name: string, title: string, description: string, _render: module.exports._render, fn: module.exports.fn}}
 */
export default class textual implements IChassisPlugin {
    name = "textual";
    title = "Server-side text-based transformation";

    install(context: IChassisContext, _options: any) {
        let options = _.extend(
            {
                mimetypes: {
                    "text/n3": {
                        extensions: "n3",
                    },
                },
            },
            _options
        );

        _.each(options.mimetypes, (opt, type) => {
            mimetypes.define(type, opt);
        });
        context.middleware.set(new TextualFeature(context, options));
    }
}

class TextualFeature implements IChassisMiddleware {
    name = "textual";
    context: IChassisContext;
    options: any;

    constructor(context: IChassisContext, options: any) {
        this.context = context;
        this.options = options;
    }

    /**
     * Collect a set of API responses as promise
     *
     * @param apis
     * @param options
     * @param token
     */

    fn(op: IOperation, _options: any): Function {
        let options = _.extend({ vars: {} }, this.options, _options);
        let context = op.context;
        let apis: APIRequests = _.extend({}, this.options.apis, options.apis);
        if (options.api) apis.data = options.api;
        let global_vars: any = _.extend({}, this.options.vars, _options.vars);

        context.log({
            code: "plugin:textual",
            resource: op.resource,
            reuseJWT: options.reuseJWT,
            apis: _.keys(apis),
        });

        // renderer middleware
        return (req, res, next) => {
            let jwt_token = options.reuseJWT ? req.header("authorization") : "";

            let op_ctx = APIs.getOperationContext(op, req, global_vars);

            // load and interpolate a template file

            let filename = Vars.$(options.filename, op_ctx);
            let raw_template = null;
            try {
                raw_template = fs.readFileSync(filename, "utf8");
            } catch (err) {
                context.warn({
                    code: "plugin:textual:file:missing",
                    filename: filename,
                    error: err.message,
                });
                return next();
            }
            let template = Handlebars.compile(raw_template);
            let file_mime = mimetypes.lookup(filename);

            context.log({
                code: "plugin:textual:apis",
                apis: _.keys(apis),
                filename: filename,
                mimetype: file_mime,
            });

            APIs.requests(this.context, apis, op_ctx, jwt_token)
                .then((results) => {
                    let render_context = _.extend({}, op_ctx, results);
                    let content = template(render_context);
                    res.status(200);
                    res.header("content-type", file_mime);
                    res.send(content);
                })
                .catch((err) => {
                    console.error(
                        "plugin:textual:apis:failed: %j",
                        err.message
                    );
                    // render error response
                    let _error = {
                        code: "plugin:textual:apis:failed",
                        filename: filename,
                        message: err.message,
                        error: err.message,
                    };
                    context.error(_error);
                    res.status(500);
                    res.send({
                        code: "plugin:textual:apis:failed",
                        request: req.a6s_uuid,
                    });
                });
        };
    }
}
